/*   Gtk+ version of the newt library 
 *
 *   Copyright (C) 1996-1999  Red Hat Software
 *   Copyright (C) 1999-2000  O'ksi'D 
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	 O'ksi'D <nickasil@linuxbox.com>
 *       Jean-Marc Lienher
 *       Rue de la Cheminee 1
 *       CH-2065 Savagnier
 *       Switzerland 
 */

#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>

#include "newt.h"
#include "newt_pr.h"

struct button {
    char * text;
    int compact;
    GtkWidget *widget;
    GtkWidget *parent;
};

/* gtk specific */
extern newtComponent currentExitComp;
extern GtkWidget* currentParent;
extern int FontSizeW;
extern int FontSizeH;
extern int ScreenW;
extern int ScreenH;
extern struct newtColors * globalColors;
extern GdkColor getColorByName (gchar *name);
/* end gtk */

static void buttonDrawIt(newtComponent co, int active, int pushed);

static void buttonDraw(newtComponent c);
static void buttonDestroy(newtComponent co);
static struct eventResult buttonEvent(newtComponent c,
				      struct event ev);
static void buttonPlace(newtComponent co, int newLeft, int newTop);

static struct componentOps buttonOps = {
    buttonDraw,
    buttonEvent,
    buttonDestroy,
    buttonPlace,
    newtDefaultMappedHandler,
} ;

static newtComponent createButton(int left, int row, const char * text, int compact) {
    newtComponent co;
    struct button * bu;

    co = malloc(sizeof(*co));
    bu = malloc(sizeof(struct button));
    co->data = bu;

    bu->text = strdup(text);
    bu->compact = compact;
    bu->widget = gtk_button_new_with_label(text);
    bu->parent = NULL;
    co->ops = &buttonOps;

    if (bu->compact) {
	co->height = 1;
	co->width = strlen(text) + 3;
    } else {
	co->height = 4;
	co->width = strlen(text) + 5;
    }

    co->top = row;
    co->left = left;
    co->takesFocus = 1;
    co->callback = NULL;
    co->isMapped = 0;

    return co;
}

newtComponent newtCompactButton(int left, int row, const char * text) {
    return createButton(left, row, text, 1);
}

newtComponent newtButton(int left, int row, const char * text) {
    return createButton(left, row, text, 0);
}

static void buttonDestroy(newtComponent co) {
    struct button * bu = co->data;
    
    free(bu->text);
    free(bu);
    free(co);
}

static void buttonPlace(newtComponent co, int newLeft, int newTop) {
    co->top = newTop;
    co->left = newLeft;

}

static void buttonDraw(newtComponent co) {
    buttonDrawIt(co, 0, 0);
}

static void buttonCb(newtComponent co) {
    currentExitComp = co;
    gtk_main_quit();
}

static void buttonFocusCb(newtComponent co) {
    if (co->callback)
        co->callback(co, co->callbackData);
}

static void buttonDrawIt(newtComponent co, int active, int pushed) {
    struct button * bu = co->data;
    GtkStyle *style;
	
    if (!co->isMapped) return;
    if (!bu->parent) {
    	bu->parent = currentParent;
	style = gtk_style_copy (gtk_widget_get_style (bu->widget));
	gtk_widget_set_style (bu->widget, style);
	gtk_widget_set_style (GTK_BIN(bu->widget)->child, style);
	if (bu->compact) {
	    style->bg[0] = getColorByName (globalColors->compactButtonBg);
	    style->fg[0] = getColorByName (globalColors->compactButtonFg);
	    gtk_widget_set_usize (bu->widget,
               FontSizeW  * co->width, FontSizeH * co->height);
	} else {
	    style->bg[0] = getColorByName (globalColors->buttonBg);
            style->fg[0] = getColorByName (globalColors->buttonFg);
            style->bg[1] = getColorByName (globalColors->actButtonBg);
	    style->bg[2] = getColorByName (globalColors->actButtonBg);
            style->fg[1] = getColorByName (globalColors->actButtonFg);
	    style->fg[2] = getColorByName (globalColors->actButtonFg);
            gtk_widget_set_usize (bu->widget,
	        FontSizeW * (co->width-1), FontSizeH * (co->height-2));
	}
	gtk_widget_show (bu->widget);
    	gtk_fixed_put (GTK_FIXED (bu->parent), bu->widget,
            FontSizeW * co->left, FontSizeH * co->top);
    	gtk_signal_connect_object (GTK_OBJECT (bu->widget), "clicked",
	    GTK_SIGNAL_FUNC (buttonCb), (GtkObject*) co);
    	gtk_signal_connect_object (GTK_OBJECT (bu->widget), "focus_in_event",
	    GTK_SIGNAL_FUNC (buttonFocusCb), (GtkObject*) co);
    	gtk_signal_connect_object (GTK_OBJECT (bu->widget), "enter_notify_event",
	    GTK_SIGNAL_FUNC (buttonFocusCb), (GtkObject*) co);
    }
}


static struct eventResult buttonEvent(newtComponent co,
				      struct event ev) {
    struct eventResult er;
    struct button * bu = co->data;

    if (ev.when == EV_NORMAL) {
	switch (ev.event) {
	  case EV_FOCUS:
	    er.result = ER_SWALLOWED;
	    if (co->callback)
	      co->callback(co, co->callbackData);
	    break;

	  case EV_UNFOCUS:
	    er.result = ER_SWALLOWED;
	    break;

	  case EV_KEYPRESS:
	    if (ev.u.key == ' ' || ev.u.key == '\r') {
		er.result = ER_EXITFORM;
	    } else 
		er.result = ER_IGNORED;
	    break;
          case EV_MOUSE:
              if (ev.u.mouse.type == MOUSE_BUTTON_DOWN &&
                  co->top <= ev.u.mouse.y &&
                  co->top + co->height - !bu->compact > ev.u.mouse.y &&
                  co->left <= ev.u.mouse.x &&
                  co->left + co->width - !bu->compact > ev.u.mouse.x) {
                  er.result = ER_EXITFORM;
              }
            break;

	}
    } else 
	er.result = ER_IGNORED;

    return er;
}
